from __future__ import annotations
from configparser import NoSectionError
import contextlib
import os
import git
import click


TPL_DIR = os.path.join(os.path.dirname(__file__), "cookiecutter")
PROJECT_PREFIX = "ckanext-"


@click.group()
def ckanext():
    """Generate CKAN extension"""


@ckanext.command()
@click.option(
    "-o",
    "--output-dir",
    help="Location to put the generated template.",
    default=".",
)
@click.argument("project", default=PROJECT_PREFIX)
@click.option("-d", "--use-defaults", is_flag=True)
def extended(output_dir: str, project: str, use_defaults: bool):
    """Generate empty extension files to expand CKAN."""
    from cookiecutter.main import cookiecutter

    template_loc = os.path.join(TPL_DIR, "extended")
    if use_defaults and project == PROJECT_PREFIX:
        click.secho("Project must be specified", fg="red")
        raise click.Abort

    if project == PROJECT_PREFIX:
        project = click.prompt("Extension's name")
    if not project.startswith(PROJECT_PREFIX):
        project = f"{PROJECT_PREFIX}{project}"

    defaults = {
        "author": "",
        "author_email": "",
        "project": project,
        "project_shortname": project[8:].lower().replace("-", "_"),
        "github_user_name": "",
        "description": "",
    }
    defaults["plugin_class_name"] = (
        defaults["project_shortname"].title().replace("_", "") + "Plugin"
    )
    git_config = git.GitConfigParser()

    with contextlib.suppress(NoSectionError):
        defaults["author"] = git_config.get("user", "name")

    with contextlib.suppress(NoSectionError):
        defaults["author_email"] = git_config.get("user", "email")

    if not use_defaults:
        defaults["author"] = click.prompt("Author's name", default=defaults["author"])
        defaults["author_email"] = click.prompt(
            "Author's email",
            default=defaults["author_email"],
        )
        defaults["github_user_name"] = click.prompt(
            "Your Github user or organization name", default=""
        )
        defaults["description"] = click.prompt(
            "Brief description of the project",
            default="",
        )

    cookiecutter(
        template_loc,
        no_input=True,
        extra_context=defaults,
        output_dir=output_dir,
    )

    click.echo(f"Written: {output_dir}")



import pytest

from ckanext.toolbelt.utils.collector import Collector


class TestCollector:
    def fun(self):
        pass

    @pytest.mark.parametrize(
        ("prefix", "separator", "basic_name", "prefixed_name"),
        [
            ("", "_", "xxx", "fun"),
            ("toolbelt", "_", "xxx", "toolbelt_fun"),
            ("toolbelt", "-", "xxx", "toolbelt-fun"),
        ],
    )
    def test_basic(self, prefix, separator, basic_name, prefixed_name):
        collector, getter = Collector(prefix, separator).split()

        assert getter() == {}

        collector(basic_name)(self.fun)
        assert getter() == {basic_name: self.fun}

        collector(self.fun)
        assert getter() == {basic_name: self.fun, prefixed_name: self.fun}



"""Benchmarks for {{ cookiecutter.project_shortname }} plugin.

Write benchmarks here while developing new features and remove them after the
solution is stabilized.

Benchmarks can be kept, but if you are not comparing results of different
versions of code, running benchmarks is just a waste of resources.

To run benchmark, just pass the tested function into `benchmark` fixture. If
function accepts arguments, add them as extra parameters for benchmark fixture.

Example:
    def test_something(benchmark):
        benchmark(my_function, arg1, arg2)
"""

from __future__ import annotations

from typing import Any

import pytest

from ckan.plugins import plugin_loaded


@pytest.mark.benchmark()
@pytest.mark.usefixtures("with_plugins")
def test_plugin(benchmark: Any):
    """First version of code."""
    benchmark(plugin_loaded, "{{ cookiecutter.project_shortname }}")


@pytest.mark.benchmark()
@pytest.mark.usefixtures("with_plugins")
def test_plugin_with_spaces(benchmark: Any):
    """Second version of code."""
    benchmark(plugin_loaded, " {{ cookiecutter.project_shortname }} ")



"""Tests for ckanext.{{ cookiecutter.project_shortname }}.plugin.

There is nothing to test here. But you can add tests for implementations into
this file. Or wait till some functionality is added to this plugin

Just as an example, we are testing that plugin can be loaded. This test has no
real value, because a lot of other tests will load the plugin implicitely, thus
we'd notice if something is broken anyway.
"""

import pytest

from ckan.plugins import plugin_loaded


@pytest.mark.usefixtures("with_plugins")
def test_plugin():
    """Plugin {{ cookiecutter.project_shortname }} is enabled in config."""
    assert plugin_loaded("{{ cookiecutter.project_shortname }}")



from __future__ import annotations

from typing import Any

import ckan.plugins.toolkit as tk
from ckan import types


def {{ cookiecutter.project_shortname }}_required(value: Any):
    """Verify that value is not empty."""
    if not value or value is tk.missing:
        raise tk.Invalid("Required")

    return value


def {{ cookiecutter.project_shortname }}_complex_validator(
    key: types.FlattenKey,
    data: types.FlattenDataDict,
    errors: types.FlattenErrorDict,
    context: types.Context,
):
    """Verify that value is not empty."""
    if not data[key]:
        errors[key].append("Required")
        raise tk.StopOnError



from __future__ import annotations

from typing import Any

import factory
import pytest
from pytest_factoryboy import register

from ckan.tests.factories import CKANFactory

from ckanext.{{ cookiecutter.project_shortname }}.model import Something


@pytest.fixture()
def clean_db(reset_db: Any, migrate_db_for: Any):
    """Apply plugin migrations whenever CKAN DB is cleaned."""
    reset_db()
    migrate_db_for("{{ cookiecutter.project_shortname }}")


@register
class SomethingFactory(CKANFactory):
    """Factory fixture for Something objects."""

    class Meta:
        model = Something
        action = "{{ cookiecutter.project_shortname }}_something_create"

    hello = factory.Faker("word")
    world = factory.Faker("word")
    plugin_data = factory.Faker("pydict", value_types=[str, int, bool])



"""CLI commands for {{ cookiecutter.project }}.

Commands added to `__all__` attribute are added to main CKAN CLI.

Example:
    ```sh
    ckan {{ cookiecutter.project_shortname }} --help
    ```
"""

from __future__ import annotations

import click

from ckan import model

__all__ = ["{{ cookiecutter.project_shortname }}"]


# Register CLI group. Code of the group is executed whenever subcommand of this
# group is invoked. Usually, group body remains empty. But if all subcommands
# contain identical initialization process, consider describing it inside
# group.
@click.group(short_help="{{ cookiecutter.project_shortname }} CLI.")
@click.pass_context
def {{ cookiecutter.project_shortname }}(ctx: click.Context):
    """CLI commands of {{ cookiecutter.project_shortname }} plugin."""
    ctx.meta["{{ cookiecutter.project_shortname }}"] = "{{ cookiecutter.project_shortname }}"


# Command decorated with `{{ cookiecutter.project_shortname }}.command()` decorator is
# registered inside the group. Name of the command matches name of the function
# with underscores replaced by hyphens. You can pass string into `.command()`
# to use a different name of the command.
#
# `click.argument` defines positional argument of the command. `click.option`
# defines optional flag with the specified short and long names.
#
# `click.pass_context` passes shared contex object to the command. Example
# below uses it to read data set by initialization code from group body.
@{{ cookiecutter.project_shortname }}.command()
@click.argument("name", default="{{ cookiecutter.project_shortname }}")
@click.option("-v", "--verbose", count=True, help="Increase verbosity")
@click.pass_context
def command(ctx: click.Context, name: str, verbose: int):
    """Nunc porta vulputate tellus."""
    msg = f"Hello, {name or ctx.meta['{{ cookiecutter.project_shortname }}']}"
    if verbose:
        msg += "!" * verbose

    click.echo(msg)


# Command can execute arbitrary code. If you are writing command that processes
# a lot of records, wrap iteration over records into `click.progressbar`. It
# results in progressbar that shows estimated time required to complete the command.
#
# When printing something, always use `click.echo` for plain output and
# `click.secho` for styled(colored) output. Never use built-in `print`
# function. It's allowed to use logging, but most likely, you don't need and
# `click.echo` is much better choice.
@{{ cookiecutter.project_shortname }}.command()
def count_users():
    """Iterate over users and count something."""
    q = model.Session.query(model.User)
    total = 0

    with click.progressbar(q, q.count()) as bar:
        for _user in bar:
            total += 1

    click.secho(f"Result: {click.style(total, bold=True)}!")



from __future__ import annotations

import os
from string import Template
from typing import Any

import click

option_plugin = click.option(
    "-p",
    "--plugin",
    default="",
    help=(
        "Name of the target plugin."
        + " By default `ckanext-<PLUGIN>` part of the extension is used."
    ),
)

option_write = click.option(
    "-w",
    "--write",
    is_flag=True,
    help="Write output to the expected location instead of STDOUT",
)


def safe_plugin_name(plugin: str) -> str:
    if not plugin:
        plugin = os.path.basename(os.getcwd())
        if plugin.startswith("ckanext-"):
            plugin = plugin[8:]
    return plugin


def render(tpl: str, data: dict[str, Any] | None = None) -> str:
    source = os.path.join(os.path.dirname(__file__), "templates", tpl)
    with open(source) as stream:
        return Template(stream.read()).safe_substitute(**data or {})


def template_source(name: str) -> str:
    return f"template_{name}.toml"


def ensure_root():
    ext = os.path.basename(os.getcwd())
    if not ext.startswith("ckanext-"):
        click.secho(
            "Can be executed only from the root directory of the extension",
            fg="red",
        )
        raise click.Abort


def produce(src: str, dest: str, data: dict[str, Any], write: bool):
    content = render(src, data).strip()

    if write:
        with open(dest, "w") as file:
            click.echo(content, file)
    else:
        click.echo(content)



"""Template helpers of the {{ cookiecutter.project_shortname }} plugin.

All non-private functions defined here are registered inside `tk.h` collection.
"""

from __future__ import annotations


def {{ cookiecutter.project_shortname }}_hello() -> str:
    """Greet the user.

    Returns:
        greeting with the plugin name.
    """
    return "Hello, {{ cookiecutter.project_shortname }}!"



import pytest

import ckan.plugins.toolkit as tk

import ckanext.toolbelt.utils.hierarchy as h

PARENT_FIELD = "notes"


@pytest.mark.ckan_config(h.CONFIG_CHILD_DISTANCE, 5)
@pytest.mark.ckan_config(h.CONFIG_PARENT_DISTANCE, 5)
@pytest.mark.ckan_config(h.CONFIG_PARENT_FIELD, PARENT_FIELD)
@pytest.mark.usefixtures(
    "with_plugins",
    "clean_db",
    "clean_index",
    "with_request_context",
)
class TestPackageHierarchy:
    def test_missing_package(self):
        with pytest.raises(tk.ObjectNotFound):
            h.package_hierarchy("not a real package")

    def test_orphan(self, package, user_factory):
        sysadmin = user_factory(sysadmin=True)
        node = h.package_hierarchy(package["id"], {"user": sysadmin["name"]})

        assert not node
        assert node.value["name"] == package["name"]

    def test_equality(self, package_factory, user_factory):
        sysadmin = user_factory(sysadmin=True)

        parent = package_factory()
        brother = package_factory(**{PARENT_FIELD: parent["id"]})
        sister = package_factory(**{PARENT_FIELD: parent["id"]})

        parent_hierarchy = h.package_hierarchy(parent["id"], {"user": sysadmin["name"]})
        brother_hierarchy = h.package_hierarchy(
            brother["id"],
            {"user": sysadmin["name"]},
        )
        sister_hierarchy = h.package_hierarchy(sister["id"], {"user": sysadmin["name"]})

        assert parent_hierarchy == brother_hierarchy == sister_hierarchy

    def test_family(self, package_factory, user_factory):
        sysadmin = user_factory(sysadmin=True)

        grand_parent = package_factory()
        parent = package_factory(**{PARENT_FIELD: grand_parent["id"]})

        brother = package_factory(**{PARENT_FIELD: parent["id"]})
        brother_child = package_factory(**{PARENT_FIELD: brother["id"]})

        sister = package_factory(**{PARENT_FIELD: parent["id"]})
        sister_child = package_factory(**{PARENT_FIELD: sister["id"]})

        hierarchy = h.package_hierarchy(parent["id"], {"user": sysadmin["name"]})

        assert hierarchy.value["id"] == grand_parent["id"]
        assert len(hierarchy) == 1

        ph = next(iter(hierarchy.leaves))
        assert ph.value["id"] == parent["id"]
        assert len(ph) == 2

        siblings_iter = iter(ph.leaves)

        # sister comes first because of default sorting by newest metadata
        # creation
        sh = next(siblings_iter)
        assert sh.value["id"] == sister["id"]
        assert len(sh) == 1

        sch = next(iter(sh.leaves))
        assert sch.value["id"] == sister_child["id"]
        assert len(sch) == 0

        bh = next(siblings_iter)
        assert bh.value["id"] == brother["id"]
        assert len(bh) == 1

        bch = next(iter(bh.leaves))
        assert bch.value["id"] == brother_child["id"]
        assert len(bch) == 0
